from open_elections.tools.reading import gather_files
from open_elections.tools.logging_helper import get_logger
import pandas as pd
import os
from typing import List, Mapping, Any, Tuple, Optional
import argparse
import sys

logger = get_logger(__name__)


BASE_SCHEMA_DEF = {
    'office': str,
    'district': int,
    'county': str,
    'precinct': str,
    'party': str,
    'candidate': str,
    'votes': int
}

STATE_SCHEMA_DEF_FILENAME = 'column_types.csv'


class DataFileException(Exception):
    def __init__(self, state: str, year: int, path: str):
        self.state = state
        self.year = year
        self.path = path


class FileEncodingException(DataFileException):
    def __init__(self, state: str, year: int, path: str, encoding_exception: Exception):
        super().__init__(state, year, path)
        self.encoding_exception = encoding_exception

    def __str__(self):
        return 'FileEncodingException caused by exception {}'.format(self.encoding_exception)


class FileFormatException(DataFileException):
    def __init__(self, state: str, year: int, path: str, format_exception: Exception):
        super().__init__(state, year, path)
        self.format_exception = format_exception

    def __str__(self):
        return 'FileFormatException caused by exception {}'.format(self.format_exception)


class ColumnMissingException(DataFileException):
    def __init__(self, state: str, year: int, path: str, column_name: str):
        super().__init__(state, year, path)
        self.column_name = column_name

    def __str__(self):
        return 'Column {} missing from file'.format(self.column_name)


class ValueTypeException(DataFileException):
    def __init__(self,
                 state: str,
                 year: int,
                 path: str,
                 column_name: str,
                 column_type: type,
                 values_and_line_numbers: Mapping[int, Any]):
        super().__init__(state, year, path)
        self.column_name = column_name
        self.column_type = column_type
        self.values_and_line_numbers = values_and_line_numbers

    def __str__(self):
        return 'ValueTypeException caused by values for column {} not being of type {}:\n {}'.format(
            self.column_name,
            self.column_type,
            self.values_and_line_numbers
        )


def validate_file(state: str, year: int, path: str, schema_def: Mapping[str, type]) -> List[DataFileException]:
    data, exception = read_file(state, year, path)
    if exception is not None:
        logger.debug('File {} cannot be parsed into a DataFrame'.format(path))
        return [exception]

    exceptions = []
    columns_present = []

    for column_name in schema_def.keys():
        logger.debug('Checking for column {} in file {}'.format(column_name, path))
        if column_name in data.columns:
            columns_present.append(column_name)
        else:
            logger.debug('Column {} missing from file {}'.format(column_name, path))
            exceptions.append(ColumnMissingException(state, year, path, column_name))

    for column_name in columns_present:
        logger.debug('Checking types of column {} in file {}'.format(column_name, path))
        column_type = schema_def[column_name]
        errors = check_types(data[column_name], column_type)
        if errors:
            exceptions.append(ValueTypeException(state, year, path, column_name, column_type, errors))

    return exceptions


def check_types(values: List[Any], column_type: type) -> Mapping[int, Any]:
    errors = {}
    for i, v in enumerate(values):
        try:
            if not pd.isna(v):
                column_type(v)
        except Exception:
            errors[i] = v

    return errors


def read_file(state: str, year: int, path: str) -> Tuple[Optional[pd.DataFrame], Optional[DataFileException]]:
    try:
        data = pd.read_csv(path)
        return data, None
    # handle the additional types of exeception
    except UnicodeDecodeError as e:
        return None, FileEncodingException(state, year, path, e)
    except pd.errors.ParserError as e:
        return None, FileFormatException(state, year, path, e)


def run_checks(base_dir: str, state: str, years: List[int] = None):
    schema_def = get_schema_def(base_dir)
    result = {}
    for year, dirpath, filename in gather_files(base_dir):
        if not years or year in years:
            path = os.path.join(dirpath, filename)
            result[path] = validate_file(state, year, path, schema_def[year])

    return result


def get_schema_def(base_dir: str) -> Mapping[int, Mapping[str, type]]:
    schema_def = get_base_schema_def(base_dir)
    state_schema_file = os.path.join(base_dir, STATE_SCHEMA_DEF_FILENAME)

    if os.path.exists(state_schema_file):
        df = pd.read_csv(state_schema_file)
        schema_def_cols = ['year', 'column', 'type']
        assert all(col in schema_def_cols for col in df.columns), 'schema def file must contain {}'.format(schema_def_cols)
        for record in df.to_dict('records'):
            year, column_name, column_type = record['year'], record['column'], record['type']
            resolved_column_type = _resolve_column_type(column_type)
            schema_def[year][column_name] = resolved_column_type

    return schema_def


def get_base_schema_def(base_dir: str) -> dict:
    years = set(year for year, _, _ in gather_files(base_dir))
    return {year: BASE_SCHEMA_DEF for year in years}


def _resolve_column_type(column_type: str):
    if column_type == 'string':
        return str
    elif column_type == 'integer':
        return int
    else:
        raise ValueError('Type {} is not supported'.format(column_type))


def display_exceptions(exceptions: Mapping[str, List[DataFileException]]):
    for path, exceptions in exceptions.items():
        logger.error('Showing exceptions for file {}'.format(path))
        for exception in exceptions:
            logger.error(exception)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--years', type=str)
    parser.add_argument('--state', type=str, required=True)
    parser.add_argument('--base-dir', type=str, required=True)
    args = parser.parse_args()

    try:
        years = [int(year) for year in args.years.split(',')]
    except ValueError as e:
        logger.error('--years expects a commma separated list of integer values')
        raise e

    assert os.path.exists(args.base_dir), 'The directory passed to --base-dir must exist'
    exceptions = run_checks(args.base_dir, args.state, years)
    display_exceptions(exceptions)
    if exceptions:
        logger.error('Exceptions found, exiting with non-zero error code')
        sys.exit(1)
    else:
        logger.info('Data is clean, exiting')
        sys.exit(0)



if __name__ == '__main__':
    main()



import os
from open_elections.tools.reading import StateMetadata, StateDataFormat
from open_elections.tools.logging_helper import get_logger
import pandas as pd
from typing import List, Callable, Optional
import importlib

BASE_DIR = '/Users/oscarbatori/Documents/open-elections'

logger = get_logger(__name__)


def build_state_metadata(state: str,
                         state_module_member: str = None,
                         strict: bool = True,
                         columns: List[str] = None,
                         vote_columns: List[str] = None,
                         df_transformers: List[Callable[[pd.DataFrame], pd.DataFrame]] = None,
                         row_cleaners: List[Callable[[dict], None]] = None) -> StateMetadata:
    """
    This is a factor method for state metadata that allows for a number of ways ot spcify state specific attributes:
        - they can be explicitly specified (for example in nationwide voting data we want to the same set of columns)
        - they can loaded from a module where we look for the variable named

    :param state:
    :param state_module_member: the member to try and retrieve from the state module of type StateDataFormat
    :param strict: indicates whether to fail in the case of the module not being present
    :param columns:
    :param vote_columns:
    :param df_transformers:
    :param row_cleaners:
    :return:
    """
    assert state in STATES, 'State {} not in: {}'.format(state, STATES)
    excluded_files = []

    # try_module is True and so we should try and grab attributes from the state
    try:
        module = importlib.import_module(get_state_module_path(state))
        if hasattr(module, state_module_member):
            state_data_format = getattr(module, state_module_member)
            if type(state_data_format) != StateDataFormat:
                raise TypeError('state_module_member must resolve an object of type StateDataFormat')
            else:
                excluded_files = excluded_files if not state_data_format.excluded_files else state_data_format.excluded_files
                columns = _combine_helper(columns, state_data_format.columns)
                vote_columns = _combine_helper(vote_columns, state_data_format.vote_columns)
                df_transformers = _combine_helper(df_transformers, state_data_format.df_transformers)
                # It is important the state specific row cleaners come first as they remove totally corrupt data,
                # the general row cleaning just does basic stuff like nulling out nan strings etc.
                row_cleaners = _combine_helper(state_data_format.row_cleaners, row_cleaners)
        else:
            if strict:
                raise NotImplementedError('No member {} in module {}'.format(state_module_member,
                                                                             get_state_module_path(state)))
    except ModuleNotFoundError:
        if strict:
            raise ModuleNotFoundError('You specified state_module_member {}, no such module found'.format(
                state_module_member
            ))
        else:
            pass

    return StateMetadata(get_state_dir(state),
                         state,
                         columns,
                         vote_columns,
                         df_transformers,
                         row_cleaners,
                         excluded_files)


def get_state_dir(state: str) -> str:
    return os.path.join(BASE_DIR, 'openelections-data-{}'.format(state))


def get_state_module_path(state: str) -> str:
    return 'open_elections.states.{}'.format(state)


def _combine_helper(left: Optional[list], right: Optional[list]):
    if left and not right:
        return left
    elif not left and right:
        return right
    else:
        return left + right


STATES = [
    'al',
    'ak',
    'az',
    'ar',
    'ca',
    'co',
    'ct',
    'de',
    'fl',
    'ga',
    'hi',
    'id',
    'il',
    'in',
    'ia',
    'ks',
    'ky',
    'la',
    'me',
    'ma',
    'mi',
    'mn',
    'ms',
    'mo',
    'mt',
    'ne',
    'nv',
    'nh',
    'nj',
    'nm',
    'ny',
    'nc',
    'nd',
    'oh',
    'ok',
    'or',
    'pa',
    'ri',
    'sc',
    'sd',
    'tn',
    'tx',
    'vt',
    'va',
    'wa',
    'wv',
    'wi',
    'wy',
]


from open_elections.tools import StateDataFormat

INVALID_FORMAT_FILES = [
'20141104__pa__general__precinct.csv',
'20140520__pa__primary__precinct.csv',
'20121106__pa__general__precinct.csv',
'20120424__pa__primary__precinct.csv',
'20080422__pa__primary__precinct.csv',
'20081104__pa__general__precinct.csv',
'20060516__pa__primary__precinct.csv',
'20061107__pa__general__precinct.csv',
'20000404__pa__primary__precinct.csv',
'20001107__pa__general__precinct.csv',
'20100518__pa__primary__precinct.csv',
'20101102__pa__general__precinct.csv',
'20180517__pa__primary__precinct.csv',
'20160426__pa__primary__precinct.csv',
'20020521__pa__primary__precinct.csv',
'20021105__pa__general__precinct.csv',
'20040427__pa__primary__precinct.csv',
'20041102__pa__general__precinct.csv'
]

national_precinct_dataformat = StateDataFormat(
    excluded_files=INVALID_FORMAT_FILES
)


from open_elections.tools import StateDataFormat


INVALID_VOTE_VALUES = [
    'Ballots',
    'Cast',
    'SEC',
    'Karl Tatgenhorst [L]',
    'AUD',
    'Suzanne Crocuh [R]',
    'AUD',
    'Michael Claytor [D]',
    'AUD',
    'John Schick [L]',
    'TREAS',
    'Kelly Mitchell [R]',
    'TREAS',
    'Mike Boland [D]',
    'TREAS',
    'Mike Jasper [L]',
    'SEN',
    'o',
    'o',
    '/46',
]


def remove_invalid_vote_counts(dic: dict):
    value = dic['votes']
    if value in INVALID_VOTE_VALUES:
        dic['votes'] = None


national_precinct_dataformat = StateDataFormat(
    row_cleaners=[remove_invalid_vote_counts]
)



from open_elections.tools import StateDataFormat

INVALID_VOTES_VALUES = [
'Moody', 'Hanson', 'McPherson', 'Hamlin', 'McCook', 'Day',
       'Harding', 'Brule', 'Grant', 'Hutchinson', 'Jones', 'Gregory',
       'Jerauld', 'Jackson', 'Edmunds', 'Campbell', 'Buffalo', 'Aurora',
       'Mellette', 'Haakon'
]


def remove_invalid_vote_counts(dic: dict):
    value = dic['votes']
    if type(value) == str and value in INVALID_VOTES_VALUES:
        dic['votes'] = None


national_precinct_dataformat = StateDataFormat(
    row_cleaners=[remove_invalid_vote_counts]
)



from open_elections.tools import StateDataFormat


def remove_emtpy_string_vote_counts(dic: dict):
    if dic['votes'] == ' ':
        dic['votes'] = None


national_precinct_dataformat = StateDataFormat(
    row_cleaners=[remove_emtpy_string_vote_counts]
)


from open_elections.tools import StateDataFormat
import pandas as pd


# need to exclude 20021105__wv__general__monongalia__precinct.csv
def remove_invalid_vote_counts(dic: dict):
    value = dic['votes']
    if type(value) == str and value == ' ':
        dic['votes'] = None


def rename_total_to_votes(df: pd.DataFrame):
    if 'total votes' in df.columns:
        return df.rename(columns={'total votes': 'votes'})
    else:
        return df


national_precinct_dataformat = StateDataFormat(
    excluded_files=['20021105__wv__general__monongalia__precinct.csv'],
    df_transformers=[rename_total_to_votes],
    row_cleaners=[remove_invalid_vote_counts]
)



from open_elections.tools import StateDataFormat
import pandas as pd

INVALID_VOTES_VALUES = ['#REF!', 'REP', 'DEM', 'LIB', 'GRN', 'votes']


def remove_invalid_vote_counts(dic: dict):
    value = dic['votes']
    if value in INVALID_VOTES_VALUES:
        dic['votes'] = None
    elif type(value) == str and '*' in value:
        dic['votes'] = int(value.lstrip().replace('*', ''))


def insert_missing_votes_column(df: pd.DataFrame) -> pd.DataFrame:
    if 'early_voting' in df.columns and 'election_day' in df.columns and 'votes' not in df.columns:
        return df.assign(votes=df['early_voting'] + df['election_day'])


national_precinct_dataformat = StateDataFormat(
    df_transformers=[insert_missing_votes_column],
    row_cleaners=[remove_invalid_vote_counts]
)



from datetime import datetime
import os
import pandas as pd
from typing import List, Tuple, Callable, Union, Iterable, Optional, Any
import re
from open_elections.tools.logging_helper import get_logger


logger = get_logger(__name__)


class StateMetadata:
    """
    Stores state metadata such as where the data lives and required format information for parsing data correctly.
    """
    def __init__(self,
                 source_dir: Union[None, str],
                 state: str,
                 columns: List[str],
                 vote_columns: List[str],
                 df_transformers: List[Callable[[pd.DataFrame], pd.DataFrame]] = None,
                 row_cleaners: Callable[[dict], dict] = None,
                 excluded_files: List[str] = None):
        self._source_dir = source_dir
        self.state = state
        self.columns = columns
        self.vote_columns = vote_columns
        self.df_transformers = df_transformers
        self.row_cleaners = row_cleaners
        self.excluded_files = excluded_files

    @property
    def source_dir(self):
        return self._source_dir

    def set_source_dir(self, value: str):
        self._source_dir = value


class StateDataFormat:
    """
    This class allows us to specify state specific columns, transformers, and row cleaners. The motivation is that we
    are loading both national and state level datasets. The former have a common set of fields, and the latter have
    their own specific vote tabulation buckets.

    The motivation for allowing us to optionally specify df_transformers and row_cleaners is to let us manipulate
    data in a way specific to the issues that arise in a certain state's data.
    """
    def __init__(self,
                 excluded_files: List[str] = None,
                 columns: List[str] = None,
                 vote_columns: List[str] = None,
                 df_transformers: List[Callable[[pd.DataFrame], pd.DataFrame]] = None,
                 row_cleaners: List[Callable[[dict], None]] = None):
        self.excluded_files = excluded_files
        self.columns = columns
        self.vote_columns = vote_columns
        self.df_transformers = df_transformers
        self.row_cleaners = row_cleaners


class VoteFile:
    DATE_POS = 0
    STATE_POS = 1
    ELECTION_POS = 2

    @classmethod
    def clean_column_names(cls, df: pd.DataFrame) -> pd.DataFrame:
        """
        Some columns have erroneous capitalization and trailing spaces, we remove those.
        :param df:
        :return:
        """
        temp = df.copy()

        return temp.rename(columns={col: col.rstrip().lower() for col in temp.columns})

    def __init__(self,
                 filepath: str,
                 state_metadata: 'StateMetadata',
                 year: int,
                 date: datetime,
                 election: str,
                 is_special: bool,
                 excluded: bool):
        self.filepath = filepath
        self.state_metadata = state_metadata
        self.date = date
        self.election = election
        self.year = year
        self.is_special = is_special
        self.excluded = excluded

        if state_metadata.df_transformers:
            self.df_transformers = [self.clean_column_names] + state_metadata.df_transformers
        else:
            self.df_transformers = [self.clean_column_names]

    def to_enriched_df(self) -> pd.DataFrame:
        logger.info('Parsing file {}'.format(self.filepath))
        try:
            df = pd.read_csv(self.filepath)
        except (pd.errors.ParserError, UnicodeDecodeError) as e:
            logger.error(str(e))
            return pd.DataFrame()
        deduplicated = df.drop_duplicates(keep='first')
        # Add some columns that we extracted from the filepath, and the filepath for debugging
        enriched_df = deduplicated.assign(state=self.state_metadata.state.upper(),
                                          year=self.year,
                                          date=self.date,
                                          election=self.election,
                                          special=self.is_special,
                                          filepath=self.filepath)

        temp = enriched_df.copy()
        if self.df_transformers:
            for transformer in self.df_transformers:
                temp = transformer(temp)

        return temp


class PrecinctFile(VoteFile):
    pass


class CountyFile(VoteFile):
    pass


class OfficeFile(VoteFile):
    pass


VoteFileBuilder = Callable[[int, str, str, 'StateMetadata', bool], 'VoteFile']
TableDataBuilder = Callable[[pd.DataFrame, StateMetadata], List[dict]]


def files_to_table_data(state_metadata: StateMetadata,
                        vote_file_builder: VoteFileBuilder,
                        table_data_builder: TableDataBuilder) -> List[dict]:
    """
    Uses state_metadata instance to map a collection of files to VoteFile objects that can be parsed into voting data.
    The vote_file_builder specifies how to map the file paths, combined with metadata, to VoteFile instances. The
    table_data_builder specifies how to take the DataFrame produced by VoteFile instance and turn it into a dict
    that can be written to Dolt.
    :param state_metadata:
    :param vote_file_builder:
    :param table_data_builder:
    :return:
    """
    vote_file_objs = build_file_objects(state_metadata, vote_file_builder)
    raw_voting_data = pd.concat([vote_file_obj.to_enriched_df() for vote_file_obj in vote_file_objs
                                 if not vote_file_obj.excluded])
    table_data = table_data_builder(raw_voting_data, state_metadata)
    return table_data


def files_to_df(state_metadata: StateMetadata, vote_file_builder: VoteFileBuilder) -> pd.DataFrame:
    """
    Utility function for getting DataFrames for files in a state, useful for debugging.
    :param state_metadata:
    :param vote_file_builder:
    :return:
    """
    vote_file_objs = build_file_objects(state_metadata, vote_file_builder)
    return pd.concat([vote_file_obj.to_enriched_df() for vote_file_obj in vote_file_objs])


def build_file_objects(state_metadata: StateMetadata, vote_file_builder: VoteFileBuilder) -> Iterable[VoteFile]:
    """
    Traverses the base_dir attribute of the StateMetadata object and uses the vote_file_builder to map that to a
    a VoteFile. Implicitly fielders by yielding when instances are not None.
    :param state_metadata:
    :param vote_file_builder:
    :return:
    """
    files = gather_files(state_metadata.source_dir)

    logger.info(
        'Parsing filenames and extracting election metadata to combine with state metadata to build VoteFile instances'
    )
    for year, dirpath, filename in files:
        result = vote_file_builder(year, dirpath, filename, state_metadata, filename in state_metadata.excluded_files)
        if result:
            yield result


def gather_files(base_dir: str) -> List[Tuple[int, str, str]]:
    logger.info('Collecting voting data files from base directory {}'.format(base_dir))
    year_dirs = [(os.path.join(base_dir, subdir), subdir)
                 for subdir in os.listdir(base_dir) if re.match(r'\d\d\d\d', subdir)]

    for year_dir, year in year_dirs:
        for dirpath, _, filenames in os.walk(year_dir):
            for filename in filenames:
                if filename.endswith('csv'):
                    try:
                        yield int(year), dirpath, filename
                    except ValueError:
                        logger.error('Error parsing year: (year, dirpath, filename): ({}, {}, {}'.format(
                            year, dirpath, filename
                        ))


def get_coerce_to_integer(null_cases: List[Any] = None):
    def inner(value: Any) -> Optional[int]:
        if null_cases and value in null_cases:
            return None
        if type(value) == str:
            # Pandas inserts strange string forms of NaN sometimes
            if value == 'nan':
                return None
            # Some are formatted as floats and strings
            if '.' in value:
                try:
                    return int(float(value))
                except ValueError:
                    raise ValueError('{} is not a valid float'.format(value))
            else:
                # numbers such as 1,000
                if ',' in value:
                    value = value.replace(',', '')
                    # Null out if there is nothing left
                    if value == '':
                        return None
                # Null out if empty
                if value == '':
                    return None
                else:
                    return int(value)
        elif pd.isna(value):
            return None
        elif type(value) == int:
            return value
        elif type(value) == float:
            return int(value)
        else:
            return value

    return inner



from open_elections.tools import StateDataFormat
import pandas as pd


def rename_total_to_votes(df: pd.DataFrame):
    if 'total_votes' in df.columns:
        return df.rename(columns={'total_votes': 'votes'})
    else:
        return df


national_precinct_dataformat = StateDataFormat(
    df_transformers=[rename_total_to_votes]
)



import pytest


def pytest_addoption(parser):
    parser.addoption('--years', type=str)
    parser.addoption('--state', type=str, required=True)
    parser.addoption('--base-dir', type=str, required=True)


@pytest.fixture
def base_dir(request):
    return request.config.getoption('base_dir')


@pytest.fixture
def state(request):
    return request.config.getoption('state')


@pytest.fixture
def years(request):
    return [int(year) for year in request.config.getoption('years').split(',')]


def _parameterize_helper(metafunc, param_name, param_value):
    if param_name in metafunc.fixturenames:
        metafunc.parametrize(param_name, param_value)



from open_elections.tools import StateDataFormat
import pandas as pd


def remove_invalid_vote_counts(dic: dict):
    value = dic['votes']
    if type(value) == str and value in ('***', '(< 25)'):
        dic['votes'] = None


def rename_votes_col(df: pd.DataFrame) -> pd.DataFrame:
    if 'total_votes' in df.columns:
        return df.rename(columns={'total_votes': 'votes'})
    else:
        return df


national_precinct_dataformat = StateDataFormat(
    df_transformers=[rename_votes_col],
    row_cleaners=[remove_invalid_vote_counts]
)



from open_elections.tools.reading import StateMetadata, VoteFile, VoteFileBuilder, build_file_objects
from open_elections.tools.logging_helper import get_logger
from typing import List, Tuple, Optional, Any, Union, Callable, Iterable
import pandas as pd

logger = get_logger(__name__)


class VoteFileIntegrityReport:
    """
    This class performs file integrity checks. It takes a PrecinctFile instance that corresponds to the file to be
    checked, and a StateMetadata instance that holds information on what columns to expect. It then parses the file,
    stores any errors if they occur, and then proceeds to validate the existence of types and columns.
    """
    def __init__(self, vote_file: VoteFile):
        self.vote_file = vote_file

    # check file parse separately from data
    # def check_file_parse(self):
    #     _, file_parse_exception = self.parse_file()
    #     if N
    #     return dict(filepath=self.vote_file.filepath, exception=str(file_parse_exception))

    def check_pre_cleaning(self) -> Tuple[Optional[dict], Optional[pd.DataFrame]]:
        """
        Runs the check on the VoteFile object this VoteFileIntegrity object has a pointer to. The check consists of
        parsing, and then if an exception is encountered returning the exception, otherwise
        :return:
        """
        df, file_parse_exception = self.parse_file()
        if df is not None:
            enriched_df = self.vote_file.to_enriched_df()
            return None, self._check_helper(enriched_df.to_dict('records'))

        return dict(filepath=self.vote_file.filepath, exception=str(file_parse_exception)), None

    def check_post_cleaning(self,
                            table_data_builder: Callable[[pd.DataFrame, StateMetadata], List[dict]]) -> pd.DataFrame:
        enriched_df = self.vote_file.to_enriched_df()
        if enriched_df.empty:
            return pd.DataFrame()
        return self._check_helper(table_data_builder(enriched_df, self.vote_file.state_metadata))

    def _check_helper(self, data: List[dict]):
        results = []
        missing_cols = set()
        for i, dic in enumerate(data):
            for col in self.vote_file.state_metadata.vote_columns:
                if col in dic and not self.is_numeric(dic[col]):
                    results.append(dict(line_number=i+1,
                                        column_name=col,
                                        present=True,
                                        type_check=False,
                                        value=dic[col] if col in dic else None))
                elif col not in dic:
                    missing_cols.add(col)
                else:
                    pass

        for col in missing_cols:
            results.append(dict(line_number=-1,
                                column_name=col,
                                present=False,
                                type_check=False,
                                value=None))
        report_df = pd.DataFrame(results).assign(state=self.vote_file.state_metadata.state,
                                                 filename=self.vote_file.filepath).drop_duplicates()
        return report_df

    @classmethod
    def is_numeric(cls, value: Any) -> bool:
        """
        This checks whether a type can be converted a integer value for the purpose of parsing. It is intentionally
        quite strict as it should flag any data where the data entry has been performed incorrectly.
        :param value:
        :return:
        """
        vt = type(value)
        if vt in (int, float) or pd.isna(value):
            return True
        elif vt == str:
            if str.isnumeric(value) or str.isnumeric(value.replace(',', '')):
                return True
            try:
                float(value)
                return True
            except ValueError:
                return False

        return False

    def parse_file(self) -> Tuple[Optional[pd.DataFrame], Optional[Exception]]:
        try:
            df = pd.read_csv(self.vote_file.filepath)
            return df, None
        except Exception as e:
            return None, e


def check_pre_clean(state_or_states: Union[StateMetadata, List[StateMetadata]],
                    vote_file_builder: VoteFileBuilder,) -> Tuple[Optional[pd.DataFrame], Optional[pd.DataFrame]]:
    if type(state_or_states) == list:
        states = state_or_states
    else:
        states = [state_or_states]

    vote_file_reports = build_vote_file_reports(states, vote_file_builder)
    reports = [vote_file_report.check_pre_cleaning() for vote_file_report in vote_file_reports]
    file_parse_reports = [file_parse_report for file_parse_report, _ in reports if file_parse_report]
    pre_clean_reports = [pre_clean_report for _, pre_clean_report in reports if pre_clean_report is not None]
    return pd.DataFrame(file_parse_reports), pd.concat(pre_clean_reports)


def check_post_clean(state_or_states: Union[StateMetadata, List[StateMetadata]],
                     vote_file_builder: VoteFileBuilder,
                     table_data_builder: Callable[[pd.DataFrame], List[dict]]) -> pd.DataFrame:
    if type(state_or_states) == list:
        states = state_or_states
    else:
        states = [state_or_states]

    vote_file_reports = build_vote_file_reports(states, vote_file_builder)
    reports = [vote_file_report.check_post_cleaning(table_data_builder) for vote_file_report in vote_file_reports
               if not vote_file_report.vote_file.excluded]
    return pd.concat(reports)


def build_vote_file_reports(states: List[StateMetadata],
                            vote_file_builder: VoteFileBuilder) -> Iterable[VoteFileIntegrityReport]:
    for state_metadata in states:
        for vote_file in build_file_objects(state_metadata, vote_file_builder):
            yield VoteFileIntegrityReport(vote_file)



from doltpy.core import Dolt
from doltpy.core.write import import_list
from typing import List
from open_elections.tools.reading import StateMetadata, VoteFileBuilder, TableDataBuilder, files_to_table_data
from open_elections.tools.logging_helper import get_logger

logger = get_logger(__name__)


def load_to_dolt(repo: Dolt,
                 dolt_table: str,
                 dolt_pks: List[str],
                 state_metadata: StateMetadata,
                 vote_file_builder: VoteFileBuilder,
                 table_data_builder: TableDataBuilder):
    """
    Load to the dolt dir/table specified using given columns for primary keys.
    :param repo:
    :param dolt_table:
    :param dolt_pks:
    :param state_metadata:
    :param vote_file_builder:
    :param table_data_builder:
    :return:
    """
    logger.info('''Loading data for state {}:
                - dolt_dir    : {}
                - dolt_table  : {}
                - dolt_pks    : {}   
            '''.format(state_metadata.state, repo.repo_dir(), dolt_table, dolt_pks))
    table_data = files_to_table_data(state_metadata, vote_file_builder, table_data_builder)
    import_list(repo, dolt_table, table_data, dolt_pks, import_mode='update', batch_size=100000)



from open_elections.tools import StateDataFormat
import pandas as pd


INVALID_VOTE_VALUES = ['S', 'Michael L Conroy', 'Kenneth P La Valle', 'Blank', 'Void',
       'Scattering', 'John J Flanagan', 'Joseph Lombardi',
       'Thomas D Croci', 'Adrienne Esposito', 'John R Alberts',
       'Philip M Boyle', 'Georgina Bowman', 'Bruce P Kennedy Jr',
       'Carl L Marcellino', 'David W Denenberg', 'Michael Venditto',
       'Bridget M Fleming', 'Kenneth P Lavalle', 'Errol D Toulon Jr',
       'Francis T Genco', 'Lee M Zeldin', 'Rick Montano',
       'David B Wright', 'Charles J Fuschillo Jr', 'Carol A Gordon']


def remove_invalid_vote_counts(dic: dict):
    value = dic['votes']
    if value in INVALID_VOTE_VALUES:
        dic['votes'] = None


def fix_vote_counts(df: pd.DataFrame) -> pd.DataFrame:
    if 'election_day' in df.columns and 'absentee' in df.columns and 'votes' not in df.columns:
        return df.assign(votes=df['election_day'] + df['absentee'])
    else:
        return df


national_precinct_dataformat = StateDataFormat(
    df_transformers=[fix_vote_counts],
    row_cleaners=[remove_invalid_vote_counts]
)



from open_elections.tools.reading import PrecinctFile, StateMetadata, get_coerce_to_integer
from open_elections.tools.config import build_state_metadata
from open_elections.validation.integrity_report_tools import check_post_clean, check_pre_clean
from open_elections.tools.logging_helper import get_logger
from open_elections.dolt.tools import load_to_dolt
from doltpy.core import Dolt
import os
from typing import List, Union, Any
import pandas as pd
from datetime import datetime
import argparse
import re


logger = get_logger(__name__)


VOTING_DATA_PKS = [
    'state',
    'year',
    'date',
    'election',
    'special',
    'office',
    'district',
    'county',
    'precinct',
    'party',
    'candidate',
]

DEFAULT_PK_VALUE = 'NA'

STATE_DATA_FORMAT_MEMBER = 'national_precinct_dataformat'


def filepath_to_precinct_file(year: int,
                              path: str,
                              file_name: str,
                              state_metadata: StateMetadata,
                              excluded: bool) -> Union[None, PrecinctFile]:
    split = file_name.split('.')[0].split('__')

    special = 'special' in split
    if special:
        split.remove('special')

    if 'precinct' not in split and 'ward' not in split:
        logger.warning(
            'Passed non-precinct file to precinct vote file builder, ignoring: '.format(os.path.join(path, file_name))
        )
        return None

    # Deals with the case of files formatted like:
    #   20160913__ny__republican__primary__richmond__precinct.csv
    # since it will have election type 'primary' and party is populated, this is redundant
    if 'primary' in split:
        if 'republican' in split:
            split.remove('republican')
        if 'democrat' in split:
            split.remove('democrat')

    state = split[PrecinctFile.STATE_POS]
    date_str = split[PrecinctFile.DATE_POS]
    election = split[PrecinctFile.ELECTION_POS]

    assert state.lower() == state_metadata.state.lower(), 'Extracted state and state_metadata.state must be the same'
    
    if date_str.startswith('_'):
        date = datetime.strptime(date_str.lstrip('_'), '%Y%m%d')
    else:
        date = datetime.strptime(date_str, '%Y%m%d')

    return PrecinctFile(os.path.join(path, file_name), state_metadata, year, date, election, special, excluded)


# TODO
#   we actually just want to run a series of row cleaners that exist per state
def extract_precinct_voting_data(raw_precinct_data: pd.DataFrame, state_metadata: StateMetadata) -> List[dict]:
    clean_precincts = raw_precinct_data.assign(precinct=raw_precinct_data['precinct'].apply(coerce_to_string),
                                               district=raw_precinct_data['district'].apply(coerce_to_string))
    not_null_pk = ensure_pks_non_null(clean_precincts[VOTING_DATA_PKS + ['votes']])
    deduplicated = not_null_pk.drop_duplicates(subset=VOTING_DATA_PKS)
    logger.warning(
        'There are {} records in the raw precinct data, and {} after de-duplicating'.format(
            len(not_null_pk),
            len(deduplicated)
        )
    )

    dicts = deduplicated.to_dict('records')

    result = []
    # We want to run coerce_votes_numeric last since it throws an exception on invalid data.
    for dic in dicts:
        for row_cleaner in state_metadata.row_cleaners:
            row_cleaner(dic)

        result.append(dic)

    return result


def coerce_votes_numeric(dic: dict):
    """
    This function takes care of some common Panda Poo that gets stuck to the data, and also handles coercing valid
    numerics to an integer object in Python memory. It also handles common strings that we infer to mean null/zero, such
    as the empty string.
    """
    dic['votes'] = get_coerce_to_integer()(dic['votes'])


def ensure_pks_non_null(raw_data: pd.DataFrame) -> pd.DataFrame:
    temp = raw_data.copy()
    for col in VOTING_DATA_PKS:
        if col in temp.columns:
            temp[col] = temp[col].fillna(DEFAULT_PK_VALUE)
        else:
            temp[col] = 'NA'

    return temp


def coerce_to_string(value: Any):
    if type(value) == int:
        return str(value)
    elif type(value) == float:
        return str(int(value))
    elif type(value) == str:
        if re.match(r'\d+\.\d+$', value):
            return str(int(float(value)))
        return value
    else:
        raise ValueError('Invalid preinct value {}'.format(value))


def clean_vote_col_names(precinct_df: pd.DataFrame) -> pd.DataFrame:
    return precinct_df.rename(columns={'vote': 'votes'}) if 'vote' in precinct_df.columns else precinct_df


def pre_clean_integrity_report(state_or_states: Union[str, List[str]]):
    if type(state_or_states) == list:
        states = state_or_states
    else:
        states = [state_or_states]

    state_metadata_list = [build_metadata_helper(state) for state in states]

    logger.info('Building state metadata')
    return check_pre_clean(state_metadata_list, filepath_to_precinct_file)


def post_clean_integrity_report(state_or_states: Union[str, List[str]]):
    if type(state_or_states) == list:
        states = state_or_states
    else:
        states = [state_or_states]

    state_metadata_list = [build_metadata_helper(state) for state in states]
    return check_post_clean(state_metadata_list, filepath_to_precinct_file, extract_precinct_voting_data)


def build_metadata_helper(state: str) -> StateMetadata:
    return build_state_metadata(state,
                                STATE_DATA_FORMAT_MEMBER,
                                False,
                                columns=VOTING_DATA_PKS + ['votes'],
                                vote_columns=['votes'],
                                df_transformers=[clean_vote_col_names, ensure_pks_non_null],
                                row_cleaners=[coerce_votes_numeric])


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--state', type=str, help='State to load data for', required=True)
    parser.add_argument('--load-data', )
    parser.add_argument('--dolt-dir', type=str, help='Dolt repo directory')
    parser.add_argument('--start-dolt-server', action='store_true')
    args = parser.parse_args()

    repo = Dolt(args.dolt_dir)
    if args.start_dolt_server:
        logger.info('start-dolt-server detected, starting server sub process')
        repo.sql_server(loglevel='trace')

    state_metadata = build_metadata_helper(args.state)

    load_to_dolt(repo,
                 'national_voting_data',
                 VOTING_DATA_PKS,
                 state_metadata,
                 filepath_to_precinct_file,
                 extract_precinct_voting_data)


if __name__ == '__main__':
    main()



from open_elections.tools import StateDataFormat


def remove_invalid_vote_counts(dic: dict):
    value = dic['votes']
    if value in ['70S', '96S', '60S', '54S', '53S', '1 1', 'Ballots Cast', 'Votes']:
        dic['votes'] = None


national_precinct_dataformat = StateDataFormat(
    row_cleaners=[remove_invalid_vote_counts]
)



from open_elections.tools import StateDataFormat


INVALID_FILES = [
    '20140826__az__primary__greenlee__precinct.csv',
    '20140826__az__primary__coconino__precinct.csv',
    '20140826__az__primary__gila__precinct.csv',
    '20140826__az__primary__yuma__precinct.csv',
    '20140826__az__primary__pinal__precinct.csv',
    '20140826__az__primary__maricopa__precinct.csv',
    '20141104__az__general__yuma__precinct.csv',
    '20140826__az__primary__mohave__precinct.csv',
    '20140826__az__primary__apache__precinct.csv',
    '20141104__az__general__gila__precinct.csv',
    '20141104__az__general__pinal__precinct.csv',
    '20141104__az__general__santa_cruz__precinct.csv',
    '20141104__az__general__coconino__precinct.csv',
    '20141104__az__general__greenlee__precinct.csv',
    '20141104__az__general__pima__precinct.csv',
    '20141104__az__general__maricopa__precinct.csv',
    '20140826__az__primary__pima__precinct.csv',
    '20141104__az__general__la_paz__precinct.csv',
    '20141104__az__general__mohave__precinct.csv',
    '20141104__az__general__apache__precinct.csv',
    '20160830__az__primary__graham__precinct.csv',
    '20160830__az__primary__santa_cruz__precinct.csv',
    '20160830__az__primary__la_paz__precinct.csv',
    '20160322__az__primary__president__la_paz__precinct.csv',
    '20160322__az__primary__president__apache__precinct.csv',
    '20160830__az__primary__apache__precinct.csv',
    '20160322__az__primary__president__cochise__precinct.csv',
    '20160830__az__primary__mohave__precinct.csv',
    '20160322__az__primary__president__mohave__precinct.csv',
    '20160830__az__primary__yuma__precinct.csv',
    '20160322__az__primary__president__pinal__precinct.csv',
    '20160322__az__primary__president__gila__precinct.csv',
    '20160830__az__primary__gila__precinct.csv',
    '20160322__az__primary__president__yuma__precinct.csv',
    '20160830__az__primary__navajo__precinct.csv',
    '20160830__az__primary__yavapai__precinct.csv',
    '20160322__az__primary__president__coconino__precinct.csv',
    '20160830__az__primary__pima__precinct.csv',
    '20160322__az__primary__president__greenlee__precinct.csv',
    '20160830__az__primary__pinal__precinct.csv',
    '20160322__az__primary__president__maricopa__precinct.csv',
    '20160322__az__primary__president__pima__precinct.csv',
    '20160322__az__primary__president__santa_cruz__precinct.csv',
    '20160830__az__primary__maricopa__precinct.csv',
    '20160830__az__primary__cochise__precinct.csv',
    '20160830__az__primary__coconino__precinct.csv',
    '20160830__az__primary__greenlee__precinct.csv'
]


national_precinct_dataformat = StateDataFormat(
    excluded_files=INVALID_FILES
)


import logging

LOG_LEVEL = logging.INFO
HANDLERS = []

logging.basicConfig(level=LOG_LEVEL,
                    format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
                    datefmt='%m-%d %H:%M:%S')


def get_logger(name: str):
    """
    Returns a logger with the given name
    :param name:
    :return:
    """
    logger = logging.getLogger(name)

    for handler in HANDLERS:
        logger.addHandler(handler)

    logger.setLevel(LOG_LEVEL)
    return logger



from open_elections.tools import StateDataFormat
import pandas as pd


def remove_invalid_vote_counts(dic: dict):
    value = dic['votes']
    if value in ['Write-ins', 'ESTES R', 'I']:
        dic['votes'] = None


def rename_vote_cols(df: pd.DataFrame) -> pd.DataFrame:
    if 'Unnamed: 6'.lower() in df.columns:
        return df.rename(columns={'Unnamed: 6'.lower(): 'votes'})
    else:
        return df


national_precinct_dataformat = StateDataFormat(
    df_transformers=[rename_vote_cols],
    row_cleaners=[remove_invalid_vote_counts]
)



from open_elections.tools import StateDataFormat


def remove_invalid_vote_counts(dic: dict):
    value = dic['votes']
    if value in ['x', 'X', 'X ', ' ', 'X394', 'X.', 'X:', '"X"', '-', '`']:
        dic['votes'] = None


national_precinct_dataformat = StateDataFormat(
    row_cleaners=[remove_invalid_vote_counts]
)



from setuptools import setup, find_packages

VERSION = '0.0.1'

setup(name='open-elections',
      version=VERSION,
      packages=find_packages(),
      install_requires=['pandas>=1.0.5',
                        'doltpy>=1.0.10'],
      tests_require=['pytest'],
      setup_requires=['wheel'],
      author='Open Elections',
      author_email='oscar@liquidata.co',
      description='A Python package for working with Open Elections data',
      entry_points={
          'console_scripts': ['validate-state=open_elections.validation.data_issues_by_state:main']
      }
  )


from open_elections.tools import StateDataFormat
import pandas as pd


def remove_invalid_vote_counts(dic: dict):
    value = dic['votes']
    if type(value) == str and value in ('Total', ''):
        dic['votes'] = None


def rename_total_to_votes(df: pd.DataFrame):
    if 'total' in df.columns:
        return df.rename(columns={'total': 'votes'})
    else:
        return df


national_precinct_dataformat = StateDataFormat(
    row_cleaners=[remove_invalid_vote_counts],
    df_transformers=[rename_total_to_votes]
)



from open_elections.validation.data_issues_by_state import run_checks, display_exceptions


def test_state_data_corruption(base_dir, state, years):
    exceptions = run_checks(base_dir, state, years)
    display_exceptions(exceptions)
    assert not exceptions



from open_elections.tools import StateDataFormat, get_coerce_to_integer
import pandas as pd


def fix_vote_counts(df: pd.DataFrame) -> pd.DataFrame:
    if 'total' in df.columns:
        return df.rename(columns={'total': 'votes'})
    else:
        breakout_cols = [col for col in ['poll', 'edr', 'abs'] if col in df.columns]
        if breakout_cols:
            temp = df.copy()
            for col in breakout_cols:
                if col in temp.columns:
                    temp[col] = temp[col].apply(get_coerce_to_integer([' -   ']))

            return temp.assign(votes=df[breakout_cols].sum(axis=1))
        else:
            return df


national_precinct_dataformat = StateDataFormat(
    df_transformers=[fix_vote_counts]
)

